"""Process your CIF to "CoRE MOF" CIF.
"""

import os, re, csv, json, glob, shutil, functools, warnings, itertools, collections
from ase.io import read, write

from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from CoREMOF.utils.atoms_definitions import METAL, COVALENTRADII
from CoREMOF.utils.ions_list import ALLIONS
import numpy as np
import pandas as pd

from ase.neighborlist import NeighborList
from scipy.sparse.csgraph import connected_components
warnings.filterwarnings('ignore')

from pymatgen.core import Structure
from pymatgen.io.ase import AseAtomsAdaptor
try:
    from mofchecker import MOFChecker
except:
    print("please run pip install git+https://github.com/sxm13/mofchecker_2.0.git@main")
from CoREMOF.utils.atoms_definitions import ATR, Coef_A, Coef_C #, BO_list, metals4check

from gemmi import cif as CIF
from PACMANCharge import pmcharge
from MOFClassifier import CLscore


def ensure_data(structure):
    """Precheck your CIF.

    Args:
        structure (str): path to your CIF.

    Returns:
        cif:
            -   added "data_struc" CIF
    """
        
    with open(structure, 'r') as file:
        lines = file.readlines()
    if not lines[1].strip().startswith('data_'):
        lines.insert(1, 'data_struc\n')
        with open(structure, 'w') as file:
            file.writelines(lines)

def ase_format(structure):
    """try to read CIF and convert to ASE format.

    Args:
        structure (str): path to your CIF.

    Returns:
        cif:
            -   ASE format CIF.
    """
        
    try:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            # mof_temp = Structure.from_file(mof,primitive=True)
            mof_temp = Structure.from_file(structure)
            mof_temp.to(filename=structure, fmt="cif")
            struc = read(structure)
            write(structure, struc)
            # print('Reading by ase: ' + mof)
    except:
        try:
            struc = read(structure)
            write(structure, struc)
            print('Reading by ase: ' + structure)
        except:
            ensure_data(structure)

class preprocess():

    """Precheck your CIF.

    Args:
        structure (str): path to your CIF.
        output_folder (str): the path to save processed CIF.

    Returns:
        Dictionary & cif:
            -   result of pre-check, has metal and carbon, has multi-structures.
            -   CIF by spliting, making primitive and making P1.
    """

    def __init__(self, structure, output_folder="result_curation"):
        self.structure = structure
        self.output = output_folder + os.sep
        os.makedirs(self.output, exist_ok=True)
        self.result_check=self.process()
                
    def process(self):
        result_check = self.split_pri_p1(self.structure, self.output)
        with open(self.output + os.path.basename(self.structure).replace(".cif","") + "_precheck.json", "w") as f:
            json.dump(result_check,f,indent=2)


    def split_pri_p1(self, structure, output_folder):
        result_check = {}
        
        structures = read(structure, index=':')
        n_struc = len(structures)
        result_check["N_structures"] = n_struc

        if n_struc > 1:
            if not os.path.exists(output_folder):
                os.makedirs(output_folder, exist_ok=True)
            print(structure, "with more than one crystal structures")
            for i, atoms in enumerate(structures):
                
                structure_= AseAtomsAdaptor.get_structure(atoms)
                sga = SpacegroupAnalyzer(structure_)
                structure_prep = sga.get_primitive_standard_structure(international_monoclinic=True, keep_site_properties=False)
                struct_name = os.path.basename(structure).replace(".cif","")+"_"+str(i+1)
                structure_prep.to(filename=os.path.join(output_folder, f"{struct_name}.cif"))

                has_metal = any(METAL.get(atom.symbol) for atom in atoms)
                has_carbon = any(atom.symbol == 'C' for atom in atoms)
                
                if has_metal:
                    if has_carbon:
                        result_check[struct_name] = "has metal and carbon"
                    else:
                        result_check[struct_name] = "missing carbon"
                else:
                    if has_carbon:
                        result_check[struct_name] = "missing metal"
                    else:
                        result_check[struct_name] = "missing metal and carbon"
        else:
            atoms = structures[0]
            struct_name = os.path.basename(structure).replace(".cif","")

            has_metal = any(METAL.get(atom.symbol) for atom in atoms)
            has_carbon = any(atom.symbol == 'C' for atom in atoms)
            
            if has_metal:
                if has_carbon:
                    result_check[struct_name] = "has metal and carbon"
                else:
                    result_check[struct_name] = "missing carbon"
            else:
                if has_carbon:
                    result_check[struct_name] = "missing metal"
                else:
                    result_check[struct_name] = "missing metal and carbon"

            structure_= AseAtomsAdaptor.get_structure(atoms)
            sga = SpacegroupAnalyzer(structure_)
            structure_prep = sga.get_primitive_standard_structure(international_monoclinic=True, keep_site_properties=False)
            temp_path = os.path.join(output_folder, f"{struct_name}.cif")
            structure_prep.to(filename=temp_path)
            
        return result_check
    

class clean():

    """Removing free solvent and coordinated solvent but keep ions based on a list.         
    
    Args:
        structure (str): path to your CIF.
        initial_skin (float): skin distance is added to the sum of vdW radii of two atoms.
        output_folder (str): the path to save processed CIF.
        saveto (bool or str): the name of csv file with clean result.

    Returns:
        CSV or cif:
            -   result of curating (name, skin, removed solvent).
            -   CIF by curating.
    """

    def __init__(self, structure, initial_skin = 0.25, output_folder="result_curation", saveto: str="clean_result.csv")-> pd.DataFrame:
        
        self.cambridge_radii = COVALENTRADII
        self.metal_list = [element for element, is_metal in METAL.items() if is_metal]
        self.ions_list = set(ALLIONS)
        
        self.structure = structure
        self.initial_skin = initial_skin
        self.output = output_folder
        os.makedirs(self.output, exist_ok=True)

        if saveto:
            self.csv_path = os.path.join(self.output, saveto)
        self.process()

    def process(self):

        """start to run curation.
        """
            
        try:
            with open(self.csv_path, mode='w', newline='') as csv_file:
                csv_writer = csv.writer(csv_file)
                csv_writer.writerow(["Name", "Skin_FSR", "Skin_ASR", "Removed_FSR", "Removed_ASR"])

                fsr_skin, fsr_results = self.run_fsr(self.structure, self.output, self.initial_skin, self.metal_list, self.ions_list)
                print(f"FSR results for {self.structure}: {fsr_results}")
                
                asr_skin, asr_results = self.run_asr(self.structure, self.output, self.initial_skin, self.metal_list, self.ions_list)
                print(f"ASR results for {self.structure}: {asr_results}")

                csv_writer.writerow([
                                        os.path.basename(self.structure),
                                        str(fsr_skin) if fsr_skin else "none",
                                        str(asr_skin) if asr_skin else "none",
                                        str(fsr_results) if fsr_results else "none",
                                        str(asr_results) if asr_results else "none"
                                    ])
        except:
            fsr_skin, fsr_results = self.run_fsr(self.structure, self.output, self.initial_skin, self.metal_list, self.ions_list)
            print(f"FSR results for {self.structure}: {fsr_results}")
            
            asr_skin, asr_results = self.run_asr(self.structure, self.output, self.initial_skin, self.metal_list, self.ions_list)
            print(f"ASR results for {self.structure}: {asr_results}")

    def run_fsr(self, mof, save_folder, initial_skin, metal_list, ions_list):

        """free solvent function.
        """

        m = mof.replace(".cif", "")
        skin = initial_skin
        try:
            while True:
                skin, printed_formulas = self.free_clean(m, save_folder, ions_list, skin)
                has_metals = False
                for e_s in printed_formulas:
                    split_formula = re.findall(r'([A-Z][a-z]?)(\d*)', e_s)
                    elements = [match[0] for match in split_formula]
                    if any(e in metal_list for e in elements):
                        has_metals = True
                        skin += 0.05
                if not has_metals:
                    break
            return skin, printed_formulas
        except Exception as e:
            print(m, str(e))


    def run_asr(self, mof, save_folder, initial_skin, metal_list, ions_list):

        """all solvent function.
        """

        m = mof.replace(".cif", "")
        skin = initial_skin
        try:
            while True:
                skin, printed_formulas = self.all_clean(m, save_folder, ions_list, skin)
                has_metals = False
                for e_s in printed_formulas:
                    split_formula = re.findall(r'([A-Z][a-z]?)(\d*)', e_s)
                    elements = [match[0] for match in split_formula]
                    if any(e in metal_list for e in elements):
                        has_metals = True
                        skin += 0.05
                if not has_metals:
                    break
            return skin, printed_formulas
        except Exception as e:
            print(f"{m} Fail: {e}")

    def build_ASE_neighborlist(self, cif, skin):

        """get list of neighbor.
        """

        radii = [self.cambridge_radii[i] for i in cif.get_chemical_symbols()]
        ASE_neighborlist = NeighborList(radii, self_interaction=False, bothways=True, skin=skin)
        ASE_neighborlist.update(cif)
        return ASE_neighborlist

    def find_clusters(self, adjacency_matrix, atom_count):

        """define cluster by the connected components of a sparse graph.
        """

        clusters = []
        cluster_count, clusterIDs = connected_components(adjacency_matrix, directed=True)
        for n in range(cluster_count):
            clusters.append([i for i in range(atom_count) if clusterIDs[i] == n])
        return clusters

    def find_metal_connected_atoms(self, structure, neighborlist):

        """get the atom connected with metal atom.
        """

        metal_connected_atoms = []
        metal_atoms = []
        for i, elem in enumerate(structure.get_chemical_symbols()):
            if elem in self.metal_list:
                neighbors, _ = neighborlist.get_neighbors(i)
                metal_connected_atoms.append(neighbors)
                metal_atoms.append(i)
        return metal_connected_atoms, metal_atoms, structure

    def CustomMatrix(self, neighborlist, atom_count):

        """convert to matrix.
        """

        matrix = np.zeros((atom_count, atom_count), dtype=int)
        for i in range(atom_count):
            neighbors, _ = neighborlist.get_neighbors(i)
            for j in neighbors:
                matrix[i][j] = 1
        return matrix

    def mod_adjacency_matrix(self, adj_matrix, MetalConAtoms, MetalAtoms, atom_count, struct):

        """modify matrix by breakdown the bond that atom connect with metal atom.
        """

        clusters = self.find_clusters(adj_matrix, atom_count)
        for i, element_1 in enumerate(MetalAtoms):
            for j, element_2 in enumerate(MetalConAtoms[i]):
                if struct[element_2].symbol == "O":
                    tmp = len(self.find_clusters(adj_matrix, atom_count))
                    adj_matrix[element_2][element_1] = 0
                    adj_matrix[element_1][element_2] = 0
                    new_clusters = self.find_clusters(adj_matrix, atom_count)
                    if tmp == len(new_clusters):
                        adj_matrix[element_2][element_1] = 1
                        adj_matrix[element_1][element_2] = 1
                    for ligand in new_clusters:
                        if ligand not in clusters:
                            tmp3 = struct[ligand].get_chemical_symbols()
                            if "O" and "H" in tmp3 and len(tmp3) == 2:
                                adj_matrix[element_2][element_1] = 1
                                adj_matrix[element_1][element_2] = 1
        return adj_matrix

    def cmp(self, x, y):
        return (x > y) - (x < y)

    def cluster_to_formula(self, cluster, cif):

        """convert to chemical formula.
        """

        symbols = [cif[i].symbol for i in cluster]
        count = collections.Counter(symbols)
        formula = ''.join([atom + (str(count[atom]) if count[atom] > 1 else '') for atom in sorted(count)])
        return formula

    def free_clean(self, input_file, save_folder, ions_list, skin):

        """workflow of removing free solvent.
        """

        try:
            print(input_file+".cif")
            cif = read(input_file+".cif")
            refcode = input_file.split("/")[-1]
            atom_count = len(cif.get_chemical_symbols())
            ASE_neighborlist = self.build_ASE_neighborlist(cif,skin)
            a = self.CustomMatrix(ASE_neighborlist,atom_count)
            b = self.find_clusters(a,atom_count)
            b.sort(key=functools.cmp_to_key(lambda x,y: self.cmp(len(x), len(y))))
            b.reverse()
            cluster_length=[]
            solvated_cluster = []
            ions_cluster = []
            printed_formulas = []
            iii=False
            print("using",skin,"as skin")
            for index, _ in enumerate(b):
                cluster_formula = self.cluster_to_formula(b[index], cif) 
                if cluster_formula in ions_list:
                    print(cluster_formula, "is ion")
                    ions_cluster.append(b[index])
                    iii=True
                else:
                    tmp = len(b[index])
                    if len(cluster_length) > 0:
                        if tmp > max(cluster_length):
                            cluster_length = []
                            solvated_cluster = []
                            solvated_cluster.append(b[index])
                            cluster_length.append(tmp)
                        elif tmp > 0.5 * max(cluster_length):
                            solvated_cluster.append(b[index])
                            cluster_length.append(tmp)
                        else:
                            formula = self.cluster_to_formula(b[index], cif)
                            if formula not in printed_formulas:
                                printed_formulas.append(formula)
                    else:
                        solvated_cluster.append(b[index])
                        cluster_length.append(tmp)
            solvated_cluster = solvated_cluster + ions_cluster
            solvated_merged = list(itertools.chain.from_iterable(solvated_cluster))
            atom_count = len(cif[solvated_merged].get_chemical_symbols())

            if iii:
                refcode=refcode.replace("_FSR","")
                new_fn = refcode + '_ION_FSR.cif'
            else:
                refcode=refcode.replace("_FSR","")
                new_fn = refcode + '_FSR.cif'
            write(os.path.join(save_folder, new_fn), cif[solvated_merged])
        
            return skin, printed_formulas
        except Exception as e:
            print(f"{input_file} Fail: {e}")

    def all_clean(self, input_file, save_folder, ions_list, skin):

        """workflow of removing all solvent.
        """

        try:
            fn = input_file
            cif = read(input_file+".cif")
            refcode = fn.split("/")[-1]
            atom_count = len(cif.get_chemical_symbols())
            ASE_neighborlist = self.build_ASE_neighborlist(cif,skin)
            a = self.CustomMatrix(ASE_neighborlist,atom_count)
            b = self.find_clusters(a,atom_count)
            b.sort(key=functools.cmp_to_key(lambda x,y: self.cmp(len(x), len(y))))
            b.reverse()
            cluster_length=[]
            solvated_cluster = []

            printed_formulas = []

            ions_cluster = []
            iii=False
            print("using",skin,"as skin")

            for index, _ in enumerate(b):
                
                cluster_formula = self.cluster_to_formula(b[index], cif) 
                if cluster_formula in ions_list:
                    print(cluster_formula, "is ion")
                    ions_cluster.append(b[index])
                    solvated_cluster.append(b[index])
                    iii=True

                else:
                    tmp = len(b[index])
                    if len(cluster_length) > 0:
                        if tmp > max(cluster_length):
                            cluster_length = []
                            solvated_cluster = []
                            solvated_cluster.append(b[index])
                            cluster_length.append(tmp)
                        if tmp > 0.5 * max(cluster_length):
                            solvated_cluster.append(b[index])
                            cluster_length.append(tmp)
                        else:
                            formula = self.cluster_to_formula(b[index], cif)
                            if formula not in printed_formulas:
                                printed_formulas.append(formula)
                    else:
                        solvated_cluster.append(b[index])
                        cluster_length.append(tmp)
                    
            solvated_merged = list(itertools.chain.from_iterable(solvated_cluster))
            
            atom_count = len(cif[solvated_merged].get_chemical_symbols())
            
            newASE_neighborlist = self.build_ASE_neighborlist(cif[solvated_merged],skin)
            MetalCon, MetalAtoms, struct = self.find_metal_connected_atoms(cif[solvated_merged], newASE_neighborlist)
            c = self.CustomMatrix(newASE_neighborlist,atom_count)
            d = self.mod_adjacency_matrix(c, MetalCon, MetalAtoms,atom_count,struct)
            solvated_clusters2 = self.find_clusters(d,atom_count)
            solvated_clusters2.sort(key=functools.cmp_to_key(lambda x,y: self.cmp(len(x), len(y))))
            solvated_clusters2.reverse()
            cluster_length=[]
            final_clusters = []

            ions_cluster2 = []
            for index, _ in enumerate(solvated_clusters2):

                cluster_formula2 = self.cluster_to_formula(solvated_clusters2[index], struct) 
                if cluster_formula2 in ions_list:
                    final_clusters.append(solvated_clusters2[index])
                    iii=True
                else:
                    tmp = len(solvated_clusters2[index])
                    if len(cluster_length) > 0:
                        if tmp > max(cluster_length):
                            cluster_length = []
                            final_clusters = []
                            final_clusters.append(solvated_clusters2[index])
                            cluster_length.append(tmp)
                        if tmp > 0.5 * max(cluster_length):
                            final_clusters.append(solvated_clusters2[index])
                            cluster_length.append(tmp)
                        else:
                            formula = self.cluster_to_formula(solvated_clusters2[index], struct)
                            if formula not in printed_formulas:
                                printed_formulas.append(formula)
                    else:
                        final_clusters.append(solvated_clusters2[index])
                        cluster_length.append(tmp)
            if iii:
                final_clusters = final_clusters+ions_cluster2
            else:
                final_clusters = final_clusters
            final_merged = list(itertools.chain.from_iterable(final_clusters))
            tmp = struct[final_merged].get_chemical_symbols()
            tmp.sort()
            if iii:
                new_fn = refcode + '_ION_ASR.cif'
            else:
                new_fn = refcode + "_ASR.cif"
            write(os.path.join(save_folder, new_fn), struct[final_merged])
            return skin, printed_formulas
    
        except Exception as e:
            print(f"{input_file} Fail: {e}")

class mof_check():

    """Not Computation-Ready (NCR) MOFs classification.

    Args:
        structure (str): path to your CIF.
        output_folder (str): the path to save checking result.

    Returns:
        Dictionary:
            -   result of NCR classification.
    """
    
    def __init__(self, structure, output_folder="result_curation"):
        self.output = output_folder + os.sep
        self.structure = structure
        os.makedirs(self.output, exist_ok=True)
        self.check()

    def check(self):

        """run checking.
        """
        
        result_check = {}

        chen_manz_result = self.Chen_Manz(self.structure)
        mof_checker_result = self.mof_checker(self.structure)

        result_check["Chen_Manz"] = chen_manz_result
        result_check["mofchecker"] = mof_checker_result

        with open(self.output + os.path.basename(self.structure).replace(".cif","") + "_Chen_Manz_mofchecker.json", "w") as f:
            json.dump(result_check,f,indent=2)

    def Chen_Manz(self, structure):

        """checking MOF by Chen and Manz method: RSC Adv., 2019,9, 36492-36507. https://doi.org/10.1039/C9RA07327B.
        """

        try:
            
            has_problem =[]

            atoms = read(structure)
            sym = atoms.get_chemical_symbols()
            
            for a in range(len(atoms)):
                H_connected = []
                nl = []
                for b in range(len(atoms)):
                    if a == b:
                        continue
                    d = atoms.get_distance(a, b, mic = True)
                    if sym[a] == 'H':
                        if d <= (0.3 + ATR[sym[a]] + ATR[sym[b]]):
                            H_connected.append(b)
                    if d < 0.5*(ATR[sym[a]] + ATR[sym[b]]):
                        has_problem.append("overlapping")
                    if d <= (ATR[sym[a]] + ATR[sym[b]]):
                        nl.append(b)
                if sym[a] == 'C':
                    bonded_ele = [sym[e] for e in nl if sym[e] not in Coef_A]
                    if len(bonded_ele) != 0:
                        pass
                    else:
                        BO = []
                        for bidx in range(len(nl)):
                            b = nl[bidx]
                            d = atoms.get_distance(a, b, mic = True)
                            BO_ab = 10**(Coef_A[sym[b]]*d + Coef_C[sym[b]])
                            if sym[b] == 'H':
                                if BO_ab > 1.25:
                                    BO_ab = 1.25
                            BO.append(BO_ab)
                        sum_BO = sum(BO)
                        if sum_BO < 3.3:
                            has_problem.append("under_carbon")
                        elif sum_BO >= 5.5:
                            has_problem.append("over_carbon")
                if len(nl) == 0:
                    has_problem.append("isolated")

            if len(has_problem) > 0:
                return list(set(has_problem))
            else:
                return ["good"]
        except Exception as e:
            return ["unknown"]
        

    def mof_checker(self, structure):
        
        """checking MOF by mofchecker 2.0: https://github.com/Au-4/mofchecker_2.0. Ref: https://doi.org/10.1039/D5DD00109A
        """

        try:
            checker = MOFChecker.from_cif(structure)
            check_result = checker.get_mof_descriptors()

            has_problem = []
            problem_keys_true = [
                                    "has_atomic_overlaps", "has_overcoordinated_c", "has_overcoordinated_n",
                                    "has_overcoordinated_h", "has_suspicious_terminal_oxo",
                                    "has_undercoordinated_c", "has_undercoordinated_n",
                                    # "has_undercoordinated_rare_earth", "has_undercoordinated_alkali_alkaline",
                                    # "has_geometrically_exposed_metal", 
                                    "has_lone_molecule", "has_high_charges"
                                ]
            problem_keys_false = ["has_metal",
                                "has_carbon",
                                "is_porous",
                                #  "has_hydrogen"
                                ]

            for key in problem_keys_true:
                if check_result.get(key, False):
                    has_problem.append(key)

            for key in problem_keys_false:
                if not check_result.get(key, True):
                    has_problem.append(key)

            if len(has_problem) > 0:
                return list(set(has_problem))
            else:
                return ["good"]
        except Exception as e:
            return ["unknown"]


class clean_pacman():

    """Removing free solvent and coordinated solvent but keep ions based on PACMAN-charge.         
    
    Args:
        structure (str): path to your CIF.
        initial_skin (float): skin distance is added to the sum of vdW radii of two atoms.
        output_folder (str): the path to save processed CIF.
        saveto (bool or str): the name of csv file with clean result.

    Returns:
        CSV or cif:
            -   result of curating (name, skin, removed solvent, charge of solvent).
            -   CIF by curating.
    """
    def __init__(self,
                 cif_path,
                 initial_skin=0.25,
                 output_folder="test"):
        
        self.cif_path = cif_path
        self.prefix = Path(self.cif_path).stem
        self.initial_skin = initial_skin
        self.output = output_folder
        self.cambridge_radii = COVALENTRADII
        self.metal_list = [element for element, is_metal in METAL.items() if is_metal]

        os.makedirs(self.output, exist_ok=True)
        self.run_pacman()
        self.process()
        

    def run_pacman(self):
        pmcharge.predict(
            cif_file=self.cif_path,
            charge_type="DDEC6",
            digits=10,
            atom_type=True,
            neutral=False,
            keep_connect=False
        )
        src = self.cif_path.replace(".cif", "_pacman.cif")
        dst = os.path.join(self.output, os.path.basename(src))

        shutil.move(src, dst)


    def process(self):
        f_d, c_d, f_s, f_i, f_i_e, c_s, c_i, c_i_e = self.run()
        info_ = {}
        info_["cif_id"] = self.prefix
        info_["skin_free"] = float(f_d)
        info_["skin_coord"] = float(c_d)
        info_["solv_free"] = f_s
        info_["ion_free"] = f_i
        info_["ion_charge_free"] = f_i_e
        info_["solv_coord"] = c_s
        info_["ion_coord"] = c_i
        info_["ion_charge_coord"] = c_i_e

        with open(os.path.join(self.output, self.prefix+".json"), "w") as f:
            json.dump(info_, f, indent=2)

    def run(self):

        skin = self.initial_skin
        atoms = read(self.cif_path)
        charges = list(CIF.read_file(os.path.join(self.output,
        self.prefix+"_pacman.cif")).sole_block().find_loop('_atom_site_charge'))

        while True:

            fsr_results = self.free_clean(atoms, charges, skin)
            skin_free, solv_free, ion_free, ion_charge_free = fsr_results
            
            has_metals = any(
            any(e in self.metal_list for e in re.findall(r'([A-Z][a-z]?)\d*', formula))
            for formula in solv_free
            )

            if has_metals:
                skin += 0.05
            else:
                break
        
        while True:
            coor_results = self.coor_clean(atoms, charges, skin)
            
            skin_coor, solv_coor, ion_coor, ion_charge_coor = coor_results
            has_metals = any(
                            any(e in self.metal_list for e in re.findall(r'([A-Z][a-z]?)\d*', formula))
                            for formula in solv_free
                            )
            if has_metals:
                skin += 0.05
            else:
                break

        return skin_free, skin_coor, solv_free, ion_free, ion_charge_free, solv_coor, ion_coor, ion_charge_coor


    def free_clean(self, atoms, charges, skin):
        
        neighborlist = self.build_ASE_neighborlist(atoms, skin)
        matrix = self.CustomMatrix(neighborlist, len(atoms))
        clusters = sorted(self.find_clusters(matrix, len(atoms)),
                            key=lambda x: len(x), reverse=True)

        main_clus, ions, solvs, solvs_idx, solvs_form = [], [], [], [], []
        ion_forms, ion_chgs = [], []
        len_main_clus = []

        for cl in clusters:
            formula = self.cluster_to_formula(cl, atoms)
            cluster_charge = sum([float(charges[i]) for i in cl if i < len(charges)])
            if len(len_main_clus) > 0:
                if len(cl) > max(len_main_clus):
                    main_clus.append(cl)
                    len_main_clus.append(len(cl))
                elif len(cl) > 0.5 * max(len_main_clus):
                    main_clus.append(cl)
                    len_main_clus.append(len(cl))
                elif abs(cluster_charge) > 0.1 and formula in ALLIONS:
                    ions.append(cl)
                    ion_forms.append(formula)
                    ion_chgs.append(cluster_charge)
                else:
                    solvs.append(cl)
                    solvs_idx.extend(cl)
                    solvs_form.append(formula)
            else:
                main_clus.append(cl)
                len_main_clus.append(len(cl))

        free_output = list(itertools.chain.from_iterable(main_clus + ions))
        write(os.path.join(self.output, self.prefix + "_FSR.cif"), atoms[free_output])
        
        return skin, solvs_form, ion_forms, ion_chgs
        

    def coor_clean(self, atoms, charges, skin):
        
        ASE_neighborlist = self.build_ASE_neighborlist(atoms, skin)
        MetalCon, MetalAtoms, struct = self.find_metal_connected_atoms(atoms, ASE_neighborlist)
        mat = self.CustomMatrix(ASE_neighborlist, len(atoms))
        mam = self.mod_adjacency_matrix(mat, MetalCon, MetalAtoms, len(atoms), struct)
        coor_clusters = sorted(self.find_clusters(mam, len(atoms)),
                               key=lambda x: len(x), reverse=True)

        main_clus, ions, solvs = [], [], []
        ion_forms, ion_chgs = [], []
        len_main_clus = []

        for cl in coor_clusters:

            formula = self.cluster_to_formula(cl, struct)
            cluster_charge = sum([float(charges[i]) for i in cl if i < len(charges)])
            all_charges = [float(charges[i]) for i in cl if i < len(charges)]
            if len(len_main_clus) > 0:
                if len(cl) > max(len_main_clus):
                    main_clus.append(cl)
                    len_main_clus.append(len(cl))
                elif len(cl) > 0.5 * max(len_main_clus):
                    main_clus.append(cl)
                    len_main_clus.append(len(cl))
                elif abs(cluster_charge) > 0.1 and formula in ALLIONS:
                    print(cluster_charge)
                    print(struct[cl])
                    print(all_charges)
                    ions.append(cl)
                    ion_forms.append(formula)
                    ion_chgs.append(cluster_charge)
                else:
                    formula = self.cluster_to_formula(cl, struct)
                    if formula not in solvs:
                        solvs.append(formula)
            else:
                main_clus.append(cl)
                len_main_clus.append(len(cl))

        final_atoms = list(itertools.chain.from_iterable(main_clus+ions))

        write(os.path.join(self.output, self.prefix+"_ASR.cif"), struct[final_atoms])

        return skin, solvs, ion_forms, ion_chgs
    

    def build_ASE_neighborlist(self, cif, skin):
        radii = [self.cambridge_radii[i] for i in cif.get_chemical_symbols()]
        neighborlist = NeighborList(radii, self_interaction=False, bothways=True, skin=skin)
        neighborlist.update(cif)
        return neighborlist

    def find_clusters(self, adjacency_matrix, atom_count):
        _, labels = connected_components(adjacency_matrix, directed=True)
        return [[i for i in range(atom_count) if labels[i] == n] for n in set(labels)]

    def CustomMatrix(self, neighborlist, atom_count):
        mat = np.zeros((atom_count, atom_count), dtype=int)
        for i in range(atom_count):
            neighbors, _ = neighborlist.get_neighbors(i)
            for j in neighbors:
                mat[i][j] = 1
        return mat

    def cluster_to_formula(self, cluster, atoms):
        symbols = [atoms[i].symbol for i in cluster]
        count = collections.Counter(symbols)
        return ''.join([el + (str(count[el]) if count[el] > 1 else '') for el in sorted(count)])

    def find_metal_connected_atoms(self, structure, neighborlist):
        metal_connected_atoms = []
        metal_atoms = []
        for i, elem in enumerate(structure.get_chemical_symbols()):
            if elem in self.metal_list:
                neighbors, _ = neighborlist.get_neighbors(i)
                metal_connected_atoms.append(neighbors)
                metal_atoms.append(i)
        return metal_connected_atoms, metal_atoms, structure

    def mod_adjacency_matrix(self, adj_matrix, MetalConAtoms, MetalAtoms, atom_count, struct):
        clusters = self.find_clusters(adj_matrix, atom_count)
        for i, element_1 in enumerate(MetalAtoms):
            for j, element_2 in enumerate(MetalConAtoms[i]):
                if struct[element_2].symbol == "O":
                    tmp = len(self.find_clusters(adj_matrix, atom_count))
                    adj_matrix[element_2][element_1] = 0
                    adj_matrix[element_1][element_2] = 0
                    new_clusters = self.find_clusters(adj_matrix, atom_count)
                    if tmp == len(new_clusters):
                        adj_matrix[element_2][element_1] = 1
                        adj_matrix[element_1][element_2] = 1
                    for ligand in new_clusters:
                        if ligand not in clusters:
                            tmp3 = struct[ligand].get_chemical_symbols()
                            if "O" in tmp3 and "H" in tmp3 and len(tmp3) == 2:
                                adj_matrix[element_2][element_1] = 1
                                adj_matrix[element_1][element_2] = 1
        return adj_matrix

try:
    from ccdc import io
    from CoREMOF.mosaec import run
except:
    print("Before using MOSAEC to check your structure, please install CSD Python API with license")


def run_MOSAEC(cif_folder, save_path="./", max_workers=64):
    """Check MOF by Metal Oxidation State Automated Error Checker: https://github.com/uowoolab/MOSAEC. Ref: https://doi.org/10.1021/jacs.5c04914        
    
    Args:
        cif_folder (str): path to the folder including all CIFs.
        save_path (str): path to save the results.
        max_workers (int): number of parallel processes.

    Returns:
        dict:
            -   results of MOSAEC.
    """

    results = run(cif_folder, max_workers=max_workers, save_path=save_path)

    return results


def run_mofclassifier(cif_folder, save_path="./mofclassifier_results.json", model="core", batch_size=64):
    """Check MOF by MOFClassifier: https://github.com/Chung-Research-Group/MOFClassifier. Ref: https://doi.org/10.1021/jacs.5c10126        
    
    Args:
        cif_folder (str): path to the folder including all CIFs.
        save_path (str): path to save the predictions.
        model (str): the name of model used for predictions.
        batch_size (int): batch size for predicting.

    Returns:
        dict:
            -   results of MOFClassifier.
    """
    all_structures = [stuc for stuc in glob.glob(cif_folder+"/*cif")[:]]
    results = CLscore.predict_batch(root_cifs=all_structures, model=model, batch_size=batch_size)
    out = {}
    for rid, s1, s2 in results:
        out[rid] = [s1, s2]
    with open(save_path, "w", encoding="utf-8") as f:
        json.dump(out, f, indent=2, ensure_ascii=False)
    return out


from setuptools import setup, find_packages
from pathlib import Path

here = Path(__file__).parent
readme_path = here / "README.md"
long_description = readme_path.read_text(encoding="utf-8") if readme_path.exists() else ""

setup(
    name='coremof_tools',
    version='0.3.4',
    author='Guobin Zhao',
    author_email='sxmzhaogb@gmail.com',
    description='Python API for CoRE MOF DB',
    license="CC-BY-4.0",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[
        'pymatgen',
        'ase',
        'juliacall',
        'molSimplify',
        'PACMAN-charge',
        'cloudpickle',
        'matminer',
        'xgboost',
        'scikit-learn==1.3.2',
        'gemmi==0.7.0',
        'phonopy',
        'networkx',
        'selfies',
        'mendeleev',
        'requests',
        "MOFClassifier==0.1.1"
    ],
    extras_require={
        'zeopp': ['zeopp-lsmo'],
        'openbabel': ['openbabel-wheel']
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Programming Language :: Python :: 3.9',
    ],
    license_files = ("LICENSE",),
    python_requires='>=3.9, <4',
    project_urls={
        "Homepage": "https://coremof-tools.readthedocs.io/",
        "Repository": "https://github.com/sxm13/CoREMOF_tools",
        "Issues": "https://github.com/mtap-research/CoRE-MOF-Tools/issues",
        "PyPI": "https://pypi.org/project/CoREMOF-tools/",
    },
)



"""ML-predicted features.
"""

import os
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

from PACMANCharge import pmcharge

import os, warnings, shutil, joblib, requests, cloudpickle
import keras
import keras.backend as K
import pandas as pd

warnings.filterwarnings('ignore')

import numpy as np
import pickle as pkl

from CoREMOF.calculation import Zeopp
from CoREMOF.calculation.mof_features import RACs, Volume

package_directory = os.path.abspath(__file__).replace("prediction.py","")

def get_files_from_github(repo, path):

    """query files from github due to limit of uploading size by PyPi.

    Args:
        repo (str): github repository.
        path (str): the path of models.

    Returns:
        Dictionary:
            -   response of downloading.
    """
        
    url = f"https://api.github.com/repos/{repo}/contents/{path}"
    headers = {'Accept': 'application/vnd.github.v3+json'}
    response = requests.get(url, headers=headers)
    response.raise_for_status() 
    return response.json()

def download_file(url, save_path):

    """download models from github due to limit of uploading size by PyPi.

    Args:
        url (str): link of downloading file.
        save_path (str): the path to save files.

    """    

    if not os.path.exists(save_path):
        response = requests.get(url)
        response.raise_for_status()
        os.makedirs(os.path.dirname(save_path), exist_ok=True) 
        with open(save_path, 'wb') as file:
            file.write(response.content)
        print(f"Downloaded {save_path}")
    else:
        pass

# repo = 'sxm13/CoREMOF_tools'
# github_paths = [
#     'models/cp_app/ensemble_models_smallML_120_100/300',
#     'models/cp_app/ensemble_models_smallML_120_100/350',
#     'models/cp_app/ensemble_models_smallML_120_100/400',
#     'models/stability'
# ]
# local_directories = [
#     '/models/cp_app/ensemble_models_smallML_120_100/300',
#     '/models/cp_app/ensemble_models_smallML_120_100/350',
#     '/models/cp_app/ensemble_models_smallML_120_100/400',
#     '/models/stability'
# ]

# path_map = dict(zip(github_paths, local_directories))

# for github_path in github_paths:
#     files_info = get_files_from_github(repo, github_path)
#     for file_info in files_info:
#         if file_info['type'] == 'file':
#             raw_url = file_info['download_url']
#             local_path_suffix = path_map[github_path].lstrip('/')
#             local_file_path = os.path.join(package_directory, local_path_suffix, os.path.basename(file_info['path']))
#             download_file(raw_url, local_file_path)

from CoREMOF.models.cp_app.descriptors import cv_features
from CoREMOF.models.cp_app.featurizer import featurize_structure
from CoREMOF.models.cp_app.predictions import predict_Cv_ensemble_structure_multitemperatures


def pacman(structure, output_folder="result_pacman", charge_type="DDEC6", digits=10, atom_type=True, neutral=True, keep_connect=False):

    """predict partial atom charge by PACMAN Charge: https://doi.org/10.1021/acs.jctc.4c00434.

    Args:
        structure (str): path to your structure.
        output_folder (str): the path to save CIF with predicted charges.
        charge_type (str): models of DDEC6, Bader, CM5 or REPEAT.
        digits (int): number of decimal places to print for partial atomic charges. ML models were trained on a 6-digit dataset.
        atom_type (bool): keep the same partial atomic charge for the same atom types (based on the similarity of partial atomic charges up to 2 decimal places).
        neutral (bool): keep the net charge is zero. We use "mean" method to neuralize the system where the excess charges are equally distributed across all atoms.
        keep_connect (bool): retain the atomic and connection information (such as _atom_site_adp_type, bond) for the structure.

    Returns:
        Dictionary & cif:
            -   predicted PBE energy and bandgap of your structure.
            -   CIF with predicted charges.
    """  
        
    results_eb = {}
    
    if not os.path.exists(output_folder):
        os.makedirs(output_folder, exist_ok=True)
    
    name = os.path.basename(structure).replace(".cif", "")
    results_eb["Name"] = name

    try:
        pmcharge.predict(
            cif_file=structure,
            charge_type=charge_type,
            digits=digits,
            atom_type=atom_type,
            neutral=neutral,
            keep_connect=keep_connect
        )
        
        pbe, bandgap = pmcharge.Energy(cif_file=structure)
        
        results_eb["PBE Energy"] = pbe
        results_eb["Bandgap"] = bandgap
      
        shutil.move(structure.replace(".cif","_pacman.cif"),
                    os.path.join(output_folder,
                                structure.split("/")[-1].replace(".cif","_pacman.cif")))

        return results_eb
    
    except Exception as e:
        print(e)

def cp(structure, T=[300, 350, 400]):
    
    """predict heat capacity by GBR models at different temperatures: https://doi.org/10.1038/s41563-022-01374-3.

    Args:
        structure (str): path to your structure.
        T (list): the temperatures of your system.

    Returns:
        Dictionary:
            -   unit by ["unit"], always "J/g/K", "J/mol/K".
            -   predicted heat capacity of your structure.    
    """  
        
    name = os.path.basename(structure).replace(".cif", "")
    featurize_structure(structure, verbos=False, saveto="features.csv")
    
    predict_Cv_ensemble_structure_multitemperatures(
                                                    path_to_models=package_directory+"models/cp_app/ensemble_models_smallML_120_100",
                                                    structure_name=name + ".cif",
                                                    features_file="features.csv", 
                                                    FEATURES=cv_features,
                                                    temperatures=T,
                                                    save_to="cp.csv"
                                                    )
    result_ = pd.read_csv("cp.csv")
    result_cp = {}
    result_cp["unit"] = "J/g/K", "J/mol/K"

    for t in T:
        result_cp[str(t)+"_mean"] = [result_["Cv_gravimetric_"+str(t)+"_mean"].iloc[0],
                                        result_["Cv_molar_"+str(t)+"_mean"].iloc[0]]
        result_cp[str(t)+"_std"] = [result_["Cv_gravimetric_"+str(t)+"_std"].iloc[0],
                                        result_["Cv_molar_"+str(t)+"_std"].iloc[0]]

    os.remove("features.csv")
    os.remove("cp.csv")

    return result_cp


def precision(y_true, y_pred):

    """
    Computes the precision metric for binary classification.

    Precision is defined as the ratio of correctly predicted positive observations 
    to the total predicted positive observations. It is given by:

        precision = TP / (TP + FP)

    where:
        - TP (True Positives)  : Correctly predicted positive cases
        - FP (False Positives)  : Incorrectly predicted positive cases

    Args:
        y_true (tensor): Ground truth binary labels (0 or 1).
        y_pred (tensor): Predicted probabilities or binary predictions.

    Returns:
        tensor: Precision score (between 0 and 1).
    """
        
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    predicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
    precision = true_positives / (predicted_positives + K.epsilon())
    return precision


def recall(y_true, y_pred):

    """
    Computes the recall metric for binary classification.

    Recall (also known as Sensitivity or True Positive Rate) measures the proportion 
    of actual positives that are correctly identified by the model. It is given by:

        recall = TP / (TP + FN)

    where:
        - TP (True Positives)  : Correctly predicted positive cases
        - FN (False Negatives)  : Actual positive cases that were predicted as negative

    Args:
        y_true (tensor): Ground truth binary labels (0 or 1).
        y_pred (tensor): Predicted probabilities or binary predictions.

    Returns:
        tensor: Recall score (between 0 and 1).
    """

    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    possible_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
    recall = true_positives / (possible_positives + K.epsilon())
    return recall


def f1(y_true, y_pred):

    """
    Computes the F1-score for binary classification.

    The F1-score is the harmonic mean of precision and recall, providing a balanced 
    metric when the dataset is imbalanced. It is given by:

        F1 = 2 * (precision * recall) / (precision + recall)

    where:
        - Precision = TP / (TP + FP)
        - Recall    = TP / (TP + FN)

    Args:
        y_true (tensor): Ground truth binary labels (0 or 1).
        y_pred (tensor): Predicted probabilities or binary predictions.

    Returns:
        tensor: F1-score (between 0 and 1).
    """
        
    p = precision(y_true, y_pred)
    r = recall(y_true, y_pred)
    return 2 * ((p * r) / (p + r + K.epsilon()))


import tensorflow as tf
# tf.config.set_visible_devices([], 'GPU')

# os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
def stability(structure):
    
    """predict stability of MOFs: https://doi.org/10.1021/jacs.1c07217, https://doi.org/10.1021/jacs.4c05879.

    Args:
        structure (str): path to your structure.

    Returns:
        Dictionary:
            -   unit by ["unit"], always "nan, °C, nan".
            -   predicted thermal, solvent and water stabilities.     
    """  
        
    solvent_feature_names = [
                            'f-chi-0-all', 'f-chi-1-all', 'f-chi-2-all', 'f-chi-3-all', 'f-Z-0-all',
                            'f-Z-1-all', 'f-Z-2-all', 'f-Z-3-all', 'f-I-0-all', 'f-I-1-all', 'f-I-2-all',
                            'f-I-3-all', 'f-T-0-all', 'f-T-1-all', 'f-T-2-all', 'f-T-3-all', 'f-S-0-all',
                            'f-S-1-all', 'f-S-2-all', 'f-S-3-all', 'mc-chi-0-all', 'mc-chi-1-all',
                            'mc-chi-2-all', 'mc-chi-3-all', 'mc-Z-0-all', 'mc-Z-1-all', 'mc-Z-2-all',
                            'mc-Z-3-all', 'mc-I-1-all', 'mc-I-2-all', 'mc-I-3-all', 'mc-T-0-all',
                            'mc-T-1-all', 'mc-T-2-all', 'mc-T-3-all', 'mc-S-0-all', 'mc-S-1-all',
                            'mc-S-2-all', 'mc-S-3-all', 'D_mc-chi-1-all', 'D_mc-chi-2-all',
                            'D_mc-chi-3-all', 'D_mc-Z-1-all', 'D_mc-Z-2-all', 'D_mc-Z-3-all',
                            'D_mc-T-1-all', 'D_mc-T-2-all', 'D_mc-T-3-all', 'D_mc-S-1-all',
                            'D_mc-S-2-all', 'D_mc-S-3-all', 'f-lig-chi-0', 'f-lig-chi-1',
                            'f-lig-chi-2', 'f-lig-chi-3', 'f-lig-Z-0', 'f-lig-Z-1', 'f-lig-Z-2',
                            'f-lig-Z-3', 'f-lig-I-0', 'f-lig-I-1', 'f-lig-I-2', 'f-lig-I-3',
                            'f-lig-T-0', 'f-lig-T-1', 'f-lig-T-2', 'f-lig-T-3', 'f-lig-S-0',
                            'f-lig-S-1', 'f-lig-S-2', 'f-lig-S-3', 'lc-chi-0-all', 'lc-chi-1-all',
                            'lc-chi-2-all', 'lc-chi-3-all', 'lc-Z-0-all', 'lc-Z-1-all', 'lc-Z-2-all',
                            'lc-Z-3-all', 'lc-I-1-all', 'lc-I-2-all', 'lc-I-3-all', 'lc-T-0-all',
                            'lc-T-1-all', 'lc-T-2-all', 'lc-T-3-all', 'lc-S-0-all', 'lc-S-1-all',
                            'lc-S-2-all', 'lc-S-3-all', 'D_lc-chi-1-all', 'D_lc-chi-2-all',
                            'D_lc-chi-3-all', 'D_lc-Z-1-all', 'D_lc-Z-2-all', 'D_lc-Z-3-all',
                            'D_lc-T-1-all', 'D_lc-T-2-all', 'D_lc-T-3-all', 'D_lc-S-1-all',
                            'D_lc-S-2-all', 'D_lc-S-3-all', 'func-chi-0-all', 'func-chi-1-all',
                            'func-chi-2-all', 'func-chi-3-all', 'func-Z-0-all', 'func-Z-1-all',
                            'func-Z-2-all', 'func-Z-3-all', 'func-I-0-all', 'func-I-1-all',
                            'func-I-2-all', 'func-I-3-all', 'func-T-0-all', 'func-T-1-all',
                            'func-T-2-all', 'func-T-3-all', 'func-S-0-all', 'func-S-1-all',
                            'func-S-2-all', 'func-S-3-all', 'D_func-chi-1-all', 'D_func-chi-2-all',
                            'D_func-chi-3-all', 'D_func-Z-1-all', 'D_func-Z-2-all', 'D_func-Z-3-all',
                            'D_func-T-1-all', 'D_func-T-2-all', 'D_func-T-3-all', 'D_func-S-1-all',
                            'D_func-S-2-all', 'D_func-S-3-all']

    thermal_feature_names = ['f-chi-0-all', 'f-chi-1-all', 'f-chi-2-all', 'f-chi-3-all',
                            'f-Z-0-all', 'f-Z-1-all', 'f-Z-2-all', 'f-Z-3-all', 'f-I-0-all',
                            'f-I-1-all', 'f-I-2-all', 'f-I-3-all', 'f-T-0-all', 'f-T-1-all',
                            'f-T-2-all', 'f-T-3-all', 'f-S-0-all', 'f-S-1-all', 'f-S-2-all',
                            'f-S-3-all', 'mc-chi-0-all', 'mc-chi-1-all', 'mc-chi-2-all',
                            'mc-chi-3-all', 'mc-Z-0-all', 'mc-Z-1-all', 'mc-Z-2-all',
                            'mc-Z-3-all', 'mc-I-1-all', 'mc-I-2-all', 'mc-I-3-all',
                            'mc-T-0-all', 'mc-T-1-all', 'mc-T-2-all', 'mc-T-3-all',
                            'mc-S-0-all', 'mc-S-1-all', 'mc-S-2-all', 'mc-S-3-all',
                            'D_mc-chi-1-all', 'D_mc-chi-2-all', 'D_mc-chi-3-all',
                            'D_mc-Z-1-all', 'D_mc-Z-2-all', 'D_mc-Z-3-all', 'D_mc-T-1-all',
                            'D_mc-T-2-all', 'D_mc-T-3-all', 'D_mc-S-1-all', 'D_mc-S-2-all',
                            'D_mc-S-3-all', 'f-lig-chi-0', 'f-lig-chi-1', 'f-lig-chi-2',
                            'f-lig-chi-3', 'f-lig-Z-0', 'f-lig-Z-1', 'f-lig-Z-2', 'f-lig-Z-3',
                            'f-lig-I-0', 'f-lig-I-1', 'f-lig-I-2', 'f-lig-I-3', 'f-lig-T-0',
                            'f-lig-T-1', 'f-lig-T-2', 'f-lig-T-3', 'f-lig-S-0', 'f-lig-S-1',
                            'f-lig-S-2', 'f-lig-S-3', 'lc-chi-0-all', 'lc-chi-1-all', 'lc-chi-2-all',
                            'lc-chi-3-all', 'lc-Z-0-all', 'lc-Z-1-all', 'lc-Z-2-all', 'lc-Z-3-all',
                            'lc-I-1-all', 'lc-I-2-all', 'lc-I-3-all', 'lc-T-0-all', 'lc-T-1-all',
                            'lc-T-2-all', 'lc-T-3-all', 'lc-S-0-all', 'lc-S-1-all', 'lc-S-2-all',
                            'lc-S-3-all', 'D_lc-chi-1-all', 'D_lc-chi-2-all', 'D_lc-chi-3-all',
                            'D_lc-Z-1-all', 'D_lc-Z-2-all', 'D_lc-Z-3-all', 'D_lc-T-1-all',
                            'D_lc-T-2-all', 'D_lc-T-3-all', 'D_lc-S-1-all', 'D_lc-S-2-all',
                            'D_lc-S-3-all', 'func-chi-0-all', 'func-chi-1-all', 'func-chi-2-all',
                            'func-chi-3-all', 'func-Z-0-all', 'func-Z-1-all', 'func-Z-2-all',
                            'func-Z-3-all', 'func-I-0-all', 'func-I-1-all', 'func-I-2-all',
                            'func-I-3-all', 'func-T-0-all', 'func-T-1-all', 'func-T-2-all',
                            'func-T-3-all', 'func-S-0-all', 'func-S-1-all', 'func-S-2-all',
                            'func-S-3-all', 'D_func-chi-1-all', 'D_func-chi-2-all', 'D_func-chi-3-all',
                            'D_func-Z-1-all', 'D_func-Z-2-all', 'D_func-Z-3-all', 'D_func-T-1-all',
                            'D_func-T-2-all', 'D_func-T-3-all', 'D_func-S-1-all', 'D_func-S-2-all',
                            'D_func-S-3-all']

    water_feature_names = ['mc-Z-3-all', 'D_mc-Z-3-all', 'D_mc-Z-2-all',
                            'D_mc-Z-1-all', 'mc-chi-3-all', 'mc-Z-1-all',
                            'mc-Z-0-all', 'D_mc-chi-2-all', 'f-lig-Z-2',
                            "ASA",
                            'f-lig-I-0', 'func-S-1-all']

    result_stability = {}
    result_stability["unit"] = "nan, °C, nan"

    dependencies = {'precision':precision,'recall':recall,'f1':f1}
    solvent_model = keras.models.load_model(package_directory+'/models/stability/final_model_flag_few_epochs.h5', custom_objects=dependencies)
    thermal_model = keras.models.load_model(package_directory+'/models/stability/final_model_T_few_epochs.h5', custom_objects=dependencies)
    with open(package_directory+'/models/stability/solvent_scaler.pkl', 'rb') as f:
        solvent_scaler = pkl.load(f)
    with open(package_directory+'/models/stability/thermal_x_scaler.pkl', 'rb') as f:
        thermal_x_scaler = pkl.load(f)
    with open(package_directory+'/models/stability/thermal_y_scaler.pkl', 'rb') as f:
        thermal_y_scaler = pkl.load(f)
    with open(package_directory+'/models/stability/water_model.pkl', 'rb') as f:
        water_model = cloudpickle.load(f)
    with open(package_directory+'/models/stability/water_scaler.pkl', 'rb') as f:
        water_scaler = pkl.load(f)

    results_pd = Zeopp.PoreDiameter(structure)
    results_pv = Zeopp.PoreVolume(structure, 1.86, 1.86, 10000, True)
    results_sa = Zeopp.SurfaceArea(structure, 1.86, 1.86, 10000, True)
    result_v = Volume(structure)
    results_sa_1_4 = Zeopp.SurfaceArea(structure, 1.4, 1.4, 10000, True)
    result_RACs = RACs(structure)
    
    '''
        zeopp4sol in the start
        LCD,
        PLD,
        LFPD,
        GPOAV_1_86,
        GPONAV_1_86,
        GPOV_1_86,
        GSA_1_86,
        POAV_1_86,
        POAV_vol_frac_1_86,
        PONAV_1_86,
        PONAV_vol_frac_1_86,
        VPOV_1_86,
        VSA_1_86,
        cell_volume
    '''

    X_solvent = [results_pd["LCD"],
                    results_pd["PLD"],
                    results_pd["LFPD"],
                    results_pv["PV"][1],
                    results_pv["NPV"][1],
                    results_pv["PV"][1] + results_pv["NPV"][1],
                    results_sa["ASA"][2],
                    results_pv["PV"][0],
                    results_pv["VF"],
                    results_pv["NPV"][0],
                    results_pv["NVF"],
                    results_pv["VF"] + results_pv["NVF"],
                    results_sa["ASA"][1],
                    result_v["total_volume"]
                ]
    
    for fn_sol in solvent_feature_names:
        try:
            X_solvent.append(result_RACs["Metal"][fn_sol])
        except:
            try:
                X_solvent.append(result_RACs["Linker"][fn_sol])
            except:
                X_solvent.append(result_RACs["Function-group"][fn_sol])

    # X_solvent = np.array(X_solvent).reshape(-1, 1).flatten().reshape(1, -1).tolist()
    
    X_solvent = solvent_scaler.transform([X_solvent])
    solvent_model_prob = solvent_model.predict(X_solvent)
    solvent_model_prob = solvent_model_prob.flatten()
    result_stability["solvent removal probability"] = float(solvent_model_prob[0])

    '''
        zeopp4therm in the end
        LCD,
        PLD,
        LFPD,
        GPOAV_1_86,
        GPONAV_1_86,
        GPOV_1_86,
        GSA_1_86,
        POAV_1_86,
        POAV_vol_frac_1_86,
        PONAV_1_86,
        PONAV_vol_frac_1_86,
        VPOV_1_86,
        VSA_1_86,
        cell_volume
    '''

    X_thermal = []
    for fn_them in thermal_feature_names:
        try:
            X_thermal.append(result_RACs["Metal"][fn_them])
        except:
            try:
                X_thermal.append(result_RACs["Linker"][fn_them])
            except:
                X_thermal.append(result_RACs["Function-group"][fn_them])

    for zeo_them in [results_pd["LCD"],
                    results_pd["PLD"],
                    results_pd["LFPD"],
                    results_pv["PV"][1],
                    results_pv["NPV"][1],
                    results_pv["PV"][1] + results_pv["NPV"][1],
                    results_sa["ASA"][2],
                    results_pv["PV"][0],
                    results_pv["VF"],
                    results_pv["NPV"][0],
                    results_pv["NVF"],
                    results_pv["VF"] + results_pv["NVF"],
                    results_sa["ASA"][1],
                    result_v["total_volume"]
                ]:
        
        X_thermal.append(zeo_them)
    X_thermal = thermal_x_scaler.transform([X_thermal])
    thermal_model_pred = thermal_y_scaler.inverse_transform(thermal_model.predict(X_thermal))
    thermal_model_pred = np.round(thermal_model_pred, 1)
    thermal_model_pred = thermal_model_pred.flatten()
    result_stability["thermal stability"] = float(thermal_model_pred[0])

    X_water = []
    for fn_water in water_feature_names:
        try:
            X_water.append(result_RACs["Metal"][fn_water])
        except:
            try:
                X_water.append(result_RACs["Linker"][fn_water])
            except:
                try:
                    X_water.append(result_RACs["Function-group"][fn_water])
                except:
                    X_water.append(results_sa_1_4[fn_water][2])

    X_water = water_scaler.transform([X_water])

    water_model_prob = water_model.predict_proba(X_water)[:,1]
    # water_model_label = water_model.predict(X_water)
    result_stability["water probability"] = float(water_model_prob[0])

    return result_stability


if __name__ == "__main__":
    repo = 'sxm13/CoREMOF_tools'
    github_paths = [
        'models/cp_app/ensemble_models_smallML_120_100/300',
        'models/cp_app/ensemble_models_smallML_120_100/350',
        'models/cp_app/ensemble_models_smallML_120_100/400',
        'models/stability'
    ]
    local_dirs = [
        '/models/cp_app/ensemble_models_smallML_120_100/300',
        '/models/cp_app/ensemble_models_smallML_120_100/350',
        '/models/cp_app/ensemble_models_smallML_120_100/400',
        '/models/stability'
    ]
    path_map = dict(zip(github_paths, local_dirs))

    for github_path in github_paths:
        try:
            files_info = get_files_from_github(repo, github_path)
        except requests.HTTPError as e:
            print(f"Warning: could not list `{github_path}` → {e}")
            continue
        for file_info in files_info:
            if file_info.get('type') != 'file':
                continue
            raw_url = file_info['download_url']
            local_suffix = path_map[github_path].lstrip('/')
            local_file = os.path.join(package_directory,
                                      local_suffix,
                                      os.path.basename(file_info['path']))
            download_file(raw_url, local_file)